'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.PathResolver = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _ConsoleUtils = require('./lib/ConsoleUtils');

var _stringUtils = require('@beautiful-code/string-utils');

var _typeUtils = require('@beautiful-code/type-utils');

var _errors = require('./errors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/*
 * TODO: Export these functions to external modules.
 */
var Externals = {
    filterObject: function filterObject(object) {
        var result = {};
        Object.keys(object).filter(function (key) {
            return object[key] !== undefined;
        }).forEach(function (key) {
            return result[key] = object[key];
        });
        return result;
    },

    hasDuplicateKeys: function hasDuplicateKeys(object, other) {
        var result = false;
        Object.keys(object).map(function (key) {
            if (other.hasOwnProperty(key)) {
                result = true;
                return false;
            }
        });
        return result;
    },

    hasKeyAvailable: function hasKeyAvailable(obj, key, onSuccess, onError) {
        var contains = obj.hasOwnProperty(key);
        try {
            if (contains) {
                throw new _errors.DuplicateKeyError('The given object has duplicate keys \'' + key + '\'. Make sure all directories have unique key, or use scopes / aliases\'');
            } else {
                (0, _typeUtils.isFunction)(onSuccess) ? onSuccess(false) : undefined;
                return false;
            }
        } catch (e) {
            (0, _typeUtils.isFunction)(onError) ? onError(true, e) : (0, _ConsoleUtils.error)(e);
            return true;
        }
        return contains;
    },

    isFileURI: function isFileURI(uri) {
        return uri.split('/').pop().indexOf('.') > -1;
    },

    isDirectoryURI: function isDirectoryURI(uri) {
        return _path2.default.split('/').pop().indexOf('.') == -1;
    }
};

var filterObject = Externals.filterObject,
    hasDuplicateKeys = Externals.hasDuplicateKeys,
    hasKeyAvailable = Externals.hasKeyAvailable,
    isDirectoryURI = Externals.isDirectoryURI,
    isFileURI = Externals.isFileURI;

/** 
 * Class that creates a suite of methods useful for resolving paths within a project. Includes a directory resolver, 
 * which creates relative path resolvers for each directory in a given directory map. Support for resolver renaming and
 * alias generation, which formats the resolver name to reflect the alias as the root rather than the full path.
 * 
 * @author Chris Coppola <mynamereallysux@gmail.com>
 */

var PathResolver =

// Class Initialization

/** 
 * Creates an instance of PathResolver. Passes arguments to {@link PathResolver#initialize} to be processed.
 * 
 * @param {!String|!Object} rootPath|paths Projects root path or Object mapping projects directory structure.
 * @param {?Object} paths|options Object mapping projects directory structure or options object.
 * @param {?Object} options Options object.
 * 
 * @see {@link PathResolver#initialize}
 */
function PathResolver() {
    _classCallCheck(this, PathResolver);

    _initialiseProps.call(this);

    this.initialize.apply(this, arguments);
}

/** 
 * Configures an instance of path resolver. Can be used after a {@link PathResolver} instance is
 * created to defer the configuration. Passes arguments to the {@link PathResolver#initialize} function to be
 * processed and validated and sets the options instance property. Then passes on the configured
 * rootPath, paths object, and designated fileroot
 * 
 * @param {!String|!Object} rootPath|paths Projects root path or Object mapping projects directory structure.
 * @param {?Object} paths|options Object mapping projects directory structure or options object.
 * @param {?Object} options Options object.
 * 
 * @function
 * @public
 */

/**
 * @property {Number} depth How many directory levels should be processed. Set to -1 for inifinite depth.
 * @property {String} fileroot Namespace, or property name, that should contain a raw directory map object.
 * @property {String} namespace Namespace, or property name, that should contain the directory resolver.
 * @property {Object} paths Object mapping projects directory structure.
 * @property {String} resolverPrefix Unvalidated options object.
 * @property {String} rootPath Root path all resolver keys and aliases will be relative to.
 * 
 * @static
 * @const
 * @memberof PathResolver
 */


/** 
 * Creates a resolver function, relative to the given path.
 * 
 * @param {!String} rootPath Absolute path which the resolver will be relative to.
 * @returns {Function} Returns a function that returns a path relative to rootPath.
 * 
 * @example
 * let resolver = makeRelativeResolver('absolute/path')
 * let path = resolver('index.html')
 * 
 * console.log(path) // absolute/path/index.html
 * @function
 * @public
 */


// Private Methods

/** 
 * Creates and adds relative resolver function to the passed in resolver object, using the key as the object key.
 * 
 * @param {!String} key Name of the resolver function to be generated.
 * @param {!String} path Absolute path which the resolver will be relative to.
 * @param {!String} resolver Object that will contain the function.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/** 
 * Core function that creates both the directory resolver and the alias resolver and returns a combined resolver object.
 * 
 * @param {!Object} directoryResolver Object that will contain relative resolver functions.
 * @param {!Object} aliasResolver Object that will contain resolved aliases.
 * @param {!Object} options The object that will contain the function.
 * @param {!Number} options.depth How many directory levels should be processed. Set to -1 for inifinite depth.
 * @param {!String} options.paths Object mapping projects directory structure or options object.
 * @param {!String} options.rootPath Projects root path or Object mapping projects directory structure.
 * @returns {Object} Returns an object containing a directory resolver and a list of aliases
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/** 
 * Format path which is a concatenation of name, an optional parent path, and an optional root path.
 * 
 * @param {!Object} name Name of the current directory or file.
 * @param {Object} parentPath Relative path of containing directories.
 * @param {Object} rootPath Absolute path which the resolved path will be relative to.
 * @returns {String} Resolved path. 
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/** 
 * Format resolver key, name of dynamically generated function.
 * 
 * @param {!Object} name Name of the current directory or file.
 * @returns {String} Resolver key, concatenation of 'resolverPrefix' option and passed in name.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/** 
 * Format scope which is a concatenation of a key and the previous scope in kebab case.
 * 
 * @param {!Object} name Name of the current directory or file.
 * @param {Object} scope String representation of the nested path in kebab case.
 * @returns {String} Formatted scope in kebab case.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/**
 * Creates a directory resolver key, or a dynamic function name, to be applied to the directory resolver.
 * 
 * @param {?String} name Name of current directory or file.
 * @param {?String} scope String representation of the nested path in kebab case.
 * 
 * @returns {String} Name for dynamically generated resolver function.
 * 
 * @private
 * @function
 * @memberof PathResolver#
 */


/**
 * Initializes directory resolver and alias resolver and determines location. Default location for directory resolver is
 * the {@link PathResolver} instance. Default location for the alias resolver is the 'aliases' property.
 * 
 * @param {?String} key Name of current directory or file.
 * @param {?String} scope String representation of the nested path in kebab case.
 * 
 * @returns {String} Name for dynamically generated resolver function.
 * 
 * @private
 * @function
 * @memberof PathResolver#
 */


/**
 * Gets the current scope. Checks if current item has an alias scope and returns it, otherwise returns scope.
 * 
 * @param {!String} localPaths Object containing current directory level mapping.
 * @param {!String} scope A string representation of the nested path in kebab case 'this-is-kebab-case'
 * @throws {TypeError} _getLocalScope was passed no valid parameters, would have returned undefined
 * @returns {String} Scope relative to alias or same scope.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/**
 * Handles initial arguments and converts them into an objects file. Validates option properties using {@link PathResolver#_validateOptions}.
 * 
 * @param {!String|!Object} rootPath|paths Projects root path or Object mapping projects directory structure.
 * @param {?Object} paths|options Object mapping projects directory structure or options object.
 * @param {?Object} options Options object.
 * 
 * @returns {Object} Formatted and validated options object.
 * @see {@link PathResolver#_validateOptions}
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/**
 * Determines if directory object map contains an alias field.
 * 
 * @param {!Object} paths Object mapping projects directory structure.
 * 
 * @returns {Boolean} True if contains alias.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/**
 * Determines if directory object map contains a local alias field.
 * 
 * @param {!Object} paths Object mapping projects directory structure.
 * 
 * @returns {Boolean} True if contains local alias.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */


/**
* Determines if directory object map contains a local alias field.
* 
* @param {!Object} fullResolver Combined resolver returned from {@link PathResolver#_createResolver}.
* 
* @returns {Boolean} True if contains local alias.
* @see {@link PathResolver#_createResolver}
* 
* @function
* @private
* @memberof PathResolver#
*/


/**
 * Validates options object and set defaults.
 * 
 * @param {!Object} options Unvalidated options object.
 * 
 * @returns {Boolean} Validated options object.
 * 
 * @function
 * @private
 * @memberof PathResolver#
 */
;

PathResolver.defaultOptions = {
    depth: -1,
    fileroot: 'files',
    namespace: 'paths',
    paths: {},
    resolverPrefix: 'resolve',
    rootPath: _fs2.default.realpathSync(process.cwd()) };

var _initialiseProps = function _initialiseProps() {
    var _this = this;

    this.initialize = function () {
        _this.options = _this._handleArgs.apply(_this, arguments);

        var _getInitialResolvers = _this._getInitialResolvers(),
            directoryResolver = _getInitialResolvers.directoryResolver,
            aliasResolver = _getInitialResolvers.aliasResolver;

        var fullResolver = _this._createResolver(directoryResolver, aliasResolver, _this.options);
        _this._setFileMap(fullResolver);
    };

    this.makeRelativeResolver = function (rootPath) {
        return function () {
            var relativePath = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';
            return _path2.default.resolve(rootPath, relativePath);
        };
    };

    this._addToResolver = function (key, path, resolver) {
        hasKeyAvailable(resolver, key, function () {
            resolver[key] = _this.makeRelativeResolver(path);
        });
    };

    this._createResolver = function (directoryResolver, aliasResolver) {
        var _options = _this.options,
            depth = _options.depth,
            paths = _options.paths,
            rootPath = _options.rootPath;


        var _resolveLevel = function _resolveLevel(_ref) {
            var localPaths = _ref.localPaths,
                parentPath = _ref.parentPath,
                scope = _ref.scope,
                _ref$index = _ref.index,
                index = _ref$index === undefined ? 0 : _ref$index,
                _ref$ignoreDuplicates = _ref.ignoreDuplicates,
                ignoreDuplicates = _ref$ignoreDuplicates === undefined ? false : _ref$ignoreDuplicates;


            var _handleStringValue = function _handleStringValue(value) {
                return _this._formatPath(value, parentPath, rootPath);
            };

            var _handleObjectValue = function _handleObjectValue(key, value) {
                var localRootPath = _this._formatPath(key, parentPath);

                var nextScope = _this._formatScope(key, scope);
                var resolverKey = _this._getDirectoryResolverKey(key, scope);

                if (_this._hasAlias(value)) {
                    _handleObjectAlias(key, value, localRootPath);
                }

                if (_this._hasLocalAlias(value)) {
                    var localAlias = value._;
                    nextScope = _this._formatScope(localAlias, scope);
                    resolverKey = _this._getDirectoryResolverKey(localAlias, scope);
                }

                var isValidDepth = depth == -1 || index <= depth;
                var duplicatesAreNotPresent = !directoryResolver.hasOwnProperty(resolverKey);

                if (isValidDepth && duplicatesAreNotPresent) {
                    _this._addToResolver(resolverKey, localRootPath, directoryResolver);
                }

                return _resolveLevel({
                    localPaths: value,
                    parentPath: localRootPath,
                    scope: nextScope,
                    index: ++index,
                    ignoreDuplicates: true
                });
            };

            var _handleObjectAlias = function _handleObjectAlias(key, value, localRootPath) {
                var aliasKey = value._.replace('@', '');
                var aliasScope = _this._getLocalScope(value, scope);
                var aliasResolverKey = _this._getDirectoryResolverKey(aliasScope);
                var aliasPath = _this._formatPath(key, parentPath, rootPath);

                if (!aliasResolver.hasOwnProperty(key)) {
                    _this._addToResolver(aliasResolverKey, aliasPath, directoryResolver);
                    _this._addToResolver(aliasKey, aliasPath, aliasResolver);
                }

                _resolveLevel({
                    localPaths: value,
                    parentPath: localRootPath,
                    scope: aliasScope,
                    index: 0
                });
            };

            var localResolver = Object.entries(localPaths).reduce(function (resolver, _ref2) {
                var _ref3 = _slicedToArray(_ref2, 2),
                    key = _ref3[0],
                    value = _ref3[1];

                if ((0, _typeUtils.isString)(value) && isFileURI(value) && key != '_') {
                    resolver[key] = _handleStringValue(value);
                } else if ((0, _typeUtils.isObject)(value)) {
                    resolver[key] = _handleObjectValue(key, value);
                }

                return resolver;
            }, {});

            return localResolver;
        };

        return _resolveLevel({ localPaths: paths });
    };

    this._formatPath = function (name, parentPath, rootPath) {
        var formatted = void 0;
        if (parentPath && rootPath) {
            formatted = _path2.default.resolve(rootPath, parentPath, name);
        } else {
            if (parentPath || rootPath) {
                formatted = _path2.default.resolve(parentPath || rootPath, name);
            } else {
                formatted = name;
            }
        }
        return formatted;
    };

    this._formatResolverKey = function (name) {
        return (0, _stringUtils.toCamelCase)(_this.options.resolverPrefix + '-' + (0, _stringUtils.toSnakeCase)(name));
    };

    this._formatScope = function (name, scope) {
        return scope ? scope + '-' + name : name;
    };

    this._getDirectoryResolverKey = function (name, scope) {
        var result = void 0;
        if (scope && name) {
            var keyToResolve = scope + '-' + name;
            result = _this._formatResolverKey(keyToResolve);
        } else if (name) {
            result = _this._formatResolverKey(name);
        }
        return result;
    };

    this._getInitialResolvers = function () {
        var namespace = _this.options.namespace;

        var directoryResolver = void 0,
            aliasResolver = void 0;

        if (namespace) {
            _this[namespace] = {};
            _this[namespace]['aliases'] = {};
            directoryResolver = _this[namespace];
        } else {
            directoryResolver = _this;
        }

        aliasResolver = directoryResolver['aliases'] = {};

        return {
            directoryResolver: directoryResolver,
            aliasResolver: aliasResolver
        };
    };

    this._getLocalScope = function (localPaths, scope) {
        if (localPaths && _this._hasAlias(localPaths) || _this._hasLocalAlias(localPaths)) {
            var localScope = localPaths._.replace('@', '');
            return localScope;
        } else if (scope) {
            return scope;
        } else {
            throw new TypeError('_getLocalScope was passed no valid parameters, would have returned undefined');
        }
    };

    this._handleArgs = function () {
        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
        }

        var options = void 0;

        switch (args.length) {
            case 0:
                {
                    (0, _ConsoleUtils.warn)('A \'PathResolver\' object was passed no initial arguments. Initialization must be done manually by running the \'initialize\' function.');
                    options = {};
                }break;
            case 1:
                {
                    if ((0, _typeUtils.isObject)(args[0])) {
                        options = {
                            paths: args[0]
                        };
                    }
                }break;
            case 2:
                {
                    if ((0, _typeUtils.isObject)(args[0]) && (0, _typeUtils.isObject)(args[1])) {
                        if (args[1].hasOwnProperty('paths')) {
                            (0, _ConsoleUtils.warn)('PathResolver was passed a \'paths\' argument and an options object with the \'paths\' property. Will use \'paths\' argument.');
                        }
                        options = _extends({}, args[1], {
                            paths: args[0]
                        });
                    }
                    if ((0, _typeUtils.isString)(args[0]) && (0, _typeUtils.isObject)(args[1])) {
                        options = {
                            rootPath: args[0],
                            paths: args[1]
                        };
                    }
                }break;
            case 3:
                {
                    if (args[2].hasOwnProperty('rootPath')) {
                        (0, _ConsoleUtils.warn)('PathResolver was passed a \'rootPath\' argument and an options object with the \'rootPath\' property. Will use \'rootPath\' argument.');
                    }
                    if (args[2].hasOwnProperty('paths')) {
                        (0, _ConsoleUtils.warn)('PathResolver was passed a \'paths\' argument and an options object with the \'paths\' property. Will use \'paths\' argument.');
                    }
                    options = _extends({}, args[2], {
                        rootPath: args[0],
                        paths: args[1]
                    });
                }break;
            default:
                {
                    throw new _errors.InvalidArgumentsError('PathResolver accepts between 1-3 arguments, \'' + args.length + '\' found.');
                }
        }

        return _this._validateOptions(options);
    };

    this._hasAlias = function (paths) {
        return paths.hasOwnProperty('_') && paths._.indexOf('@') == 0;
    };

    this._hasLocalAlias = function (paths) {
        return paths.hasOwnProperty('_') && paths._.indexOf('@') == -1;
    };

    this._setFileMap = function (fullResolver) {
        var _options2 = _this.options,
            fileroot = _options2.fileroot,
            rootPath = _options2.rootPath;

        _this[fileroot] = fullResolver;
        _this[fileroot]['_root'] = rootPath;
    };

    this._validateOptions = function (options) {
        var namespace = options.namespace,
            fileroot = options.fileroot,
            rootPath = options.rootPath,
            resolverPrefix = options.resolverPrefix,
            other = _objectWithoutProperties(options, ['namespace', 'fileroot', 'rootPath', 'resolverPrefix']);

        var validated = Object.assign({}, PathResolver.defaultOptions, filterObject(_extends({}, other, {
            namespace: (0, _typeUtils.isString)(namespace) && namespace.trim().length > 0 ? namespace : undefined,
            fileroot: (0, _typeUtils.isString)(fileroot) && fileroot.trim().length > 0 ? fileroot : undefined,
            rootPath: (0, _typeUtils.isString)(rootPath) && rootPath.trim().length > 0 ? rootPath : undefined,
            resolverPrefix: (0, _typeUtils.isString)(rootPath) && rootPath.trim().length > 0 ? rootPath : undefined
        })));

        return validated;
    };
};

exports.default = PathResolver;
exports.PathResolver = PathResolver;